import logging

from rocolib.api.composables.graph.HyperEdge import HyperEdge
from rocolib.utils.utils import prefix as prefixString
from rocolib.utils.roco2sim.Node import ShapeNode
import rocolib.utils.numsym as np
from rocolib.utils.roco2sim.format_3d import format_wrl


log = logging.getLogger(__name__)

def inflate(face, thickness=.1, edges=False):
  dt = np.array([[0],[0],[thickness/2.],[0]])
  nf = face-dt
  pf = face+dt

  faces = []

  if edges:
    faces.append(np.transpose(np.array((pf[:,0], nf[:,0], pf[:,1]))))
    faces.append(np.transpose(np.array((nf[:,0], nf[:,1], pf[:,1]))))
  else:
    faces.append(pf)          # top face
    faces.append(nf[:,::-1])  # bottom face

  return faces


def _triangulate(faces, **kwargs):
  """Create triangulated faces in 3D space with thickness"""
  scale = .001  # roco units : mm ; STL units m
  import triangle
  import numpy as np

  triangles_coord_index = []
  triangles_facet_normal = []
  facets_index = np.empty([1, 3])
  B_add = 0
  k = 0  # index initiated

  for i, f in enumerate(faces):

    r = f[0]  # coord transformation matrix
    A = f[1]  # vertices info

    facets_coord_index = []
    facets_normal = []
    facets = []
    B = triangle.triangulate(A, opts='p')

    if not 'triangles' in B:
      log.warning("No triangles in " + f[2])
      continue

    thickness = "thickness" in kwargs and kwargs["thickness"]
    if thickness:
      for t in [np.transpose(np.array([list(B['vertices'][x]) + [0, 1] for x in (face[0], face[1], face[2])])) for
                face in B['triangles']]:
        for x in inflate(t, thickness=thickness):
          facets_coord_index.extend(np.transpose(np.dot(r, x) * scale))
          facets_normal.extend(np.transpose(np.dot(r, x) * scale))
          index_current = [[k, k + 1, k + 2]]
          k = k + 3
          facets_index = np.concatenate((facets_index, index_current), 0)
      for t in [np.transpose(np.array([list(A['vertices'][x]) + [0, 1] for x in (edge[0], edge[1])])) for edge in
                A['segments']]:
        for x in inflate(t, thickness=thickness, edges=True):
          facets_coord_index.extend(np.transpose(np.dot(r, x) * scale))
          facets_normal.extend(np.transpose(np.dot(r, x) * scale))
          index_current = [[k, k + 1, k + 2]]
          k = k + 3
          facets_index = np.concatenate((facets_index, index_current), 0)
    else:
      for t in [np.transpose(np.array([list(B['vertices'][x]) + [0, 1] for x in (face[0], face[1], face[2])])) for
                face in B['triangles']]:
        facets.append(np.dot(r, t) * scale)
        facets_coord_index.extend(np.transpose(np.dot(r, t)) * scale)
        facets_normal.extend(np.transpose(np.dot(r, t)) * scale)
        #B_triangle = B['triangles'] + B_add
        #B_add = np.amax(B_triangle) + 1
        index_current = [[k, k + 1, k + 2]]
        k = k + 3
        facets_index = np.concatenate((facets_index, index_current), 0)


    triangles_coord_index.extend(facets_coord_index)
    triangles_facet_normal.extend(facets_normal)

  facets_index_return = np.delete(facets_index, 0, 0)

  return format_wrl(triangles_coord_index, triangles_facet_normal, facets_index_return)


class Graph():
  def __init__(self):
    self.faces = []
    self.facelists = []
    self.edges = []
    self.shapeNodes=None

  def addFace(self, f, prefix=None, root=False, faceEdges=None, faceAngles=None, faceFlips=None):
    if prefix:
      f.prefix(prefix)
    if f in self.faces:
      raise ValueError("Face %s already in graph" % f.name)
    if root:
        self.faces.insert(0, f)
    else:
        self.faces.append(f)

    if faceEdges is not None:
      f.renameEdges(faceEdges, faceAngles, faceFlips, self.edges)
      if prefix:
        f.prefixEdges(prefix)

    self.rebuildEdges()
    return self

  def attachFace(self, fromFace, toFace, prefix=None, transform=None):
    self.addFace(toFace, prefix)
    self.mergeFace(fromFace, toFace.name, transform)

  def mergeFace(self, fromFaceName, toFaceName, transform=None):
    fromFace = self.getFace(fromFaceName)
    toFace = self.getFace(toFaceName)
    if transform is None:
        transform = np.eye(4)
    fromFace.addFace(toFace, transform)
    toFace.addFace(fromFace, np.inv(transform))

  def attachEdge(self, fromEdge, newFace, newEdge, prefix=None, root=False, angle=0, edgeType=None, joints=None):
    # XXX should set angle from a face, not absolute angle of the face
    self.addFace(newFace, prefix, root)

    if fromEdge is not None:
      newEdge = prefixString(prefix, newEdge)
      self.mergeEdge(fromEdge, newEdge, angle=angle, edgeType=edgeType, joints=joints)

  def delFace(self, facename):
    for (i, f) in enumerate(self.faces):
      if f.name == facename:
        f.disconnectAll()
        self.faces.pop(i)
        self.rebuildEdges()
        return self

    return self

  def getFace(self, name):
    for f in self.faces:
      if f.name == name:
        return f
    return None

  def getEdge(self, name):
    for e in self.edges:
      if e.name == name:
        return e
    return None

  def prefix(self, prefix):
    for e in self.edges:
      e.rename(prefixString(prefix, e.name))
    for f in self.faces:
      f.rename(prefixString(prefix, f.name))

  def renameEdge(self, fromname, toname):
    e = self.getEdge(fromname)
    if e:
      e.rename(toname)

  def rebuildEdges(self):
    self.edges = []
    for f in self.faces:
      for e in f.edges:
        if e not in self.edges:
          self.edges.append(e)

  def invertEdges(self):
    # swap mountain and valley folds
    for e in self.edges:
      for f in e.faces:
        e.faces[f] = (-e.faces[f][0], e.faces[f][1])
    for f in self.faces:
      f.inverted = not f.inverted

  def addTab(self, edge1, edge2, angle=0, width=10):
    self.mergeEdge(edge1, edge2, angle=angle, tabWidth=width)

  def mergeEdge(self, edge1, edge2, angle=0, tabWidth=None, edgeType=None, joints=None, swap=False):
    e1 = self.getEdge(edge1)
    e2 = self.getEdge(edge2)
    if e1 is None or e2 is None:
      logstr  = f"Edge not found trying to merge ({edge1}, {edge2}) in edges:\n  "
      logstr += "\n  ".join([e.name for e in self.edges])
      log.error(logstr)
      raise AttributeError("Edge not found")
    if swap:
      e1, e2 = e2, e1

    if len(e2.faces) > 1:
      log.warning("Adding more than two faces to an edge, currently not well supported.")
      e2.mergeWith(e1, angle=angle, flip=False, tabWidth=tabWidth)
    else:
      e2.mergeWith(e1, angle=angle, flip=True, tabWidth=tabWidth)
    self.edges.remove(e1)
    
    e2.setType(edgeType)
    if joints:
        for joint in joints.joints:
            e2.addJoint(joint)

    return self

  def splitEdge(self, edge):
    old_edge = edge
    old_edge_name = edge.name
    new_edges_and_faces = []

    for i, face in enumerate(list(old_edge.faces)):
        length = old_edge.length
        angle = old_edge.faces[face][0]
        flip = old_edge.faces[face][1]

        new_edge_name = old_edge_name + '.se' + str(i)
        new_edge = HyperEdge(new_edge_name, length)
        face.replaceEdge(old_edge, new_edge, angle, flip=False )
        new_edges_and_faces.append((new_edge_name, face, length, angle, flip))

    self.rebuildEdges()
    return new_edges_and_faces

  def jointify(self, **kwargs):
    for e in self.edges:
        if e.isNotFlat() and "joint" in kwargs:
            e.setType("JOINT")
            e.addJoint(kwargs["joint"])
            #print "jointing ", e.name

  def tabify(self, tabFace=None, tabDecoration=None, slotFace=None, slotDecoration=None, **kwargs):
    for e in self.edges:
      if e.isTab():
        #print "tabbing ", e.name
        for (edgename, face, length, angle, flip) in self.splitEdge(e):
          if flip:
            #print "-- tab on: ", edgename, face.name, angle
            if tabDecoration is not None and ((abs(angle) > 179.5 and abs(angle) < 180.5) or tabFace is None):
              tabDecoration(face, edgename, e.tabWidth, flip=True, **kwargs)
            elif tabFace is not None:
              tab = tabFace(length, e.tabWidth, **kwargs)
              self.attachEdge(edgename, tab, tab.MAINEDGE, prefix=edgename, angle=0)
          else:
            #print "-- slot on: ", edgename, face.name
            if slotFace is not None:
              # XXX TODO: set angle appropriately
              slot = slotFace(length, e.tabWidth, **kwargs)
              self.attachEdge(edgename, slot, slot.MAINEDGE, prefix=edgename, angle=0)
            if slotDecoration is not None:
              slotDecoration(face, edgename, e.tabWidth, **kwargs)

    #TODO: extend this to three+ edges
    #component.addConnectors((conn, cname), new_edges[0], new_edges[1], depth, tabattachment=None, angle=0)

  def flip(self):
    return
    for f in self.faces:
      f.flip()

  def transform(self, scale=1, angle=0, origin=(0,0)):
    pass

  def dotransform(self, scale=1, angle=0, origin=(0,0)):
    for f in self.faces:
      f.transform(scale, angle, origin)

  def mirrorY(self):
    return
    for f in self.faces:
      f.transform( mirrorY())

  def mirrorX(self):
    return
    for f in self.faces:
      f.transform( mirrorX())

  def toString(self):
    print()
    for f in self.faces:
      print(f.name + repr(f.edges))

  def graphObj(self):
    g = {}
    for f in self.faces:
      g[f.name] = dict([(e and e.name or "", e) for e in f.edges])
    return g

  def showGraph(self):
    import objgraph
    objgraph.show_refs(self.graphObj(), max_depth = 2, filter = lambda x : isinstance(x, (dict, HyperEdge)))

  def place(self, force=False, transform3D=None):
    if force:
        self.unplace()
        self.facelists = []

    if transform3D is None:
        transform3D = np.eye(4)

    while True:
        for f in self.faces:
            if f.transform2D is not None and f.transform3D is not None:
                continue
            else:
                f.place(None, None, transform3D, self.facelists)
                break
        else:
            break

    #IPython.embed()
    self.rebuildEdges()
  
  def unplace(self):
  
    for f in self.faces:
        f.transform2D = None
        f.transform3D = None
        
    for e in self.edges:
        e.pts2D = None
        e.pts3D = None

  def toShapeNodes(self, **kwargs):
    """Create a shape node list"""
    self.place()
    shapeNodes = []
    for i, facesInFacelists in enumerate(self.facelists):
      faces = []
      for f in facesInFacelists:
        if f.area > 0:
          faces.append([f.transform3D, f.getTriangleDict(), f.name])
        else:
          log.info(f"Omitting face {f.name} with area {f.area} from wbo")
      # Crate 1 shape node per set of faces.
      wrl = _triangulate(faces, **kwargs)
      shapeNodes.append(ShapeNode(f.name, wrl, f.transform3D))

    self.shapeNodes = shapeNodes

  def to3D(self, fp, **kwargs):
    """stl or webots expects fp instance from handle. if None, not creating it"""
    if self.shapeNodes is None:
      self.toShapeNodes(**kwargs)

    if kwargs.get('format') == 'stl':
      f = []
      for sn in self.shapeNodes:
        f.extend(sn.getSTL().face)
        from .stlwriter import ASCII_STL_Writer as STL_Writer
        writer = STL_Writer(fp)
        writer.add_faces(f)
        writer.close()

    if kwargs.get('format') == 'webots':
        from rocolib.utils.roco2sim.wbo_nodes import wboConverter
        fp.write(wboConverter(self.shapeNodes))


  '''
  @staticmethod
  def joinAlongEdge((g1, prefix1, edge1), (g2, prefix2, edge2), merge=True, useOrigEdge=False, angle=0):
    # TODO(mehtank): make sure that edges are congruent

    g = Graph()
    for f in g1.faces:
      g.addFace(f.copy(prefix(prefix1, f.name)), faceEdges = [prefix(prefix1, e.name) for e in f.edges])

    return g.attach(edge1, (g2, prefix2, edge2), merge=merge, useOrigEdge=useOrigEdge, angle=angle)

  @staticmethod
  def joinAlongFace((g1, prefix1, face1), (g2, prefix2, face2), toKeep=1):
    # TODO(mehtank): make sure that faces are congruent
    g = Graph()
    for f in g1.faces:
      g.addFace(f.copy(prefix1 + "." + f.name), faceEdges = [prefix1 + "." + e.name for e in f.edges])
    for f in g2.faces:
      g.addFace(f.copy(prefix2 + "." + f.name), faceEdges = [prefix2 + "." + e.name for e in f.edges])
    f1 = g.getFace(prefix1 + "." + face1)
    f2 = g.getFace(prefix2 + "." + face2)
    for (e1, e2) in zip(f1.edges, f2.edges):
      e1.mergeWith(e2)
    if toKeep < 2:
      g.delFace(f2.name)
    if toKeep < 1:
      g.delFace(f1.name)
    return g
  '''
